;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "consult-notes" "20260222.1928"
  "Manage notes with consult."
  '((emacs   "28.1")
    (consult "1.0")
    (s       "1.12.0")
    (dash    "2.19"))
  :url "https://github.com/mclear-tools/consult-notes"
  :commit "94e5b19ec84363438eb7114e59c059a67f3d7c59"
  :revdesc "94e5b19ec843"
  :keywords '("convenience")
  :authors '(("Colin McLear" . "mclear@fastmail.com")))
